"""
chouette.metrics.plugins.messages

Contains messages that Collectors and CollectorPlugins use to communicate.
"""
# pylint: disable=too-few-public-methods
from typing import Iterator

from pykka import ActorRef  # type: ignore


__all__ = ["StatsRequest", "StatsResponse"]


class StatsRequest:
    """
    StatsRequest is a message that a Collector sends to a Plugin.

    It requests that Plugin to gather stats and send them back as a
    StatsResponse message.

    To give the Plugin a chance to send its stats back, this message
    contains a `sender` ActorRef parameter.
    """

    __slots__ = ["sender"]

    def __init__(self, sender: ActorRef):
        """
        Args:
            sender: ActorRef of a Collector that requested stats.
        """
        self.sender = sender

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"<{self.__class__.__name__} from {self.sender}>"


class StatsResponse:
    """
    StatsResponse is a message that a Plugin sends to a Collector.

    It's being produces as a response on a StatsRequest message.

    Since interaction between Collectors and Plugins is completely
    non-blocking, the Collector won't know what plugin sent it a
    message, to this message contains the Plugin's name in its
    `producer` attribute.
    """

    __slots__ = ["producer", "stats"]

    def __init__(self, producer: str, stats: Iterator):
        """
        Args:
            producer: Name of a Plugin that produced the message.
            stats: Iterator over WrappedMetric objects.
        """
        self.producer = producer
        self.stats = stats

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"<{self.__class__.__name__} from {self.producer}>"



import json
import logging
from functools import reduce
from itertools import groupby
from typing import Any, List, Optional, Tuple

from chouette_iot import ChouetteConfig
from chouette_iot._singleton_actor import VitalActor
from chouette_iot.metrics import MergedMetric
from chouette_iot.metrics.wrappers import WrappersFactory
from chouette_iot.storages import RedisStorage
from chouette_iot.storages.messages import (
    CleanupOutdatedRecords,
    CollectKeys,
    CollectValues,
    DeleteRecords,
    StoreRecords,
)

__all__ = ["MetricsAggregator"]

logger = logging.getLogger("chouette-iot")


class MetricsAggregator(VitalActor):
    """
    MetricsAggregator is an actor that is responsible for data aggregation.
    https://docs.datadoghq.com/developers/dogstatsd/data_aggregation/

    Environment variable `AGGREGATE_INTERVAL` determines how often data is
    aggregated and it's analogous to a `flush interval` from DogStatsD
    documentation. Collected metrics are being processed in groups,
    where every group represents all the metrics emitted during
    `AGGREGATE_INTERVAL` seconds.

    If Chouette was working constantly, it usually don't have lots of
    raw metrics to process, it's just some amount of metrics collected
    since the last Aggregation execution.
    However, if for some reason Chouette was stopped or down and other
    applications were producing metrics, grouping thousands of keys
    into small groups, gathering lots of groups of metrics and
    processing every group can take quite a while.

    To avoid duplicating metrics in this situation, MetricsAggregator
    intentionally has a blocking workflow based on `ask` patterns.

    If one aggregate call takes more than `AGGREGATE_INTERVAL` to
    finish, other calls will be queued in the actor's mailbox and
    will be executed only when the first call is finished and processed
    metrics are cleaned up from a storage.
    """

    def __init__(self):
        super().__init__()
        config = ChouetteConfig()
        self.aggregate_interval = config.aggregate_interval
        self.metric_ttl = config.metric_ttl
        self.metrics_wrapper = WrappersFactory.get_wrapper(config.metrics_wrapper)
        if not self.metrics_wrapper:
            logger.warning(
                "[%s] No MetricsWrapper found. "
                "Raw metrics won't be collected and aggregated.",
                self.name,
            )
        self.redis = None

    def on_receive(self, message: Any) -> bool:
        """
        On any message MetricsAggregator performs the following routine:

        1. Tries to clean up outdated raw metrics, because Datadog rejects
        metrics older than 4 hours.
        2. Gets all the keys of all the metrics with their timestamps.
        3. Groups these keys based on their timestamps.
        4. For every group - fetches raw metrics from a storage.
        5. Casts them into MergedMetric objects.
        6. Processes these MergedMetrics with a specified MetricsWrapper.
        7. Stores produced WrappedMetrics to a storage.
        8. Cleans up original raw metrics from a storage.

        Args:
            message: Anything.
        Return: Whether all the raw metrics were processed and stored.
        """
        logger.debug("[%s] Cleaning up outdated raw metrics.", self.name)
        self.redis = RedisStorage.get_instance()
        self.redis.ask(
            CleanupOutdatedRecords("metrics", ttl=self.metric_ttl, wrapped=False)
        )
        if not self.metrics_wrapper:
            return True

        keys = self.redis.ask(CollectKeys("metrics", wrapped=False))
        grouped_keys = MetricsMerger.group_metric_keys(keys, self.aggregate_interval)
        logger.info(
            "[%s] Separated %s metric keys into %s groups of %s seconds.",
            self.name,
            len(keys),
            len(grouped_keys),
            self.aggregate_interval,
        )

        return all(map(self._process_metrics, grouped_keys))

    def _process_metrics(self, keys: List[bytes]) -> bool:
        """
        Processes metrics.

        1. Fetches metrics from a storage by their keys.
        2. Merges them into a list of MergedMetric objects.
        3. Casts these MergedMetrics into WrappedMetrics using logic
        of a specified MetricsWrapper.
        4. Stores produced WrappedMetrics to a storage.
        5. Removes original raw metrics from a storage.

        Args:
            keys: List of metric keys to fetch data from a storage.
        Returns: Whether metrics were processed and cleaned up.
        """
        b_records = self.redis.ask(CollectValues("metrics", keys, wrapped=False))
        merged_metrics = MetricsMerger.merge_metrics(b_records, self.aggregate_interval)
        logger.info(
            "[%s] Merged %s raw metrics into %s Merged Metrics.",
            self.name,
            len(b_records),
            len(merged_metrics),
        )
        wrapped_metrics = self.metrics_wrapper.wrap_metrics(merged_metrics)
        logger.info(
            "[%s] Wrapped %s Merged Metrics into %s Wrapped Metrics.",
            self.name,
            len(merged_metrics),
            len(wrapped_metrics),
        )
        # Storing processed messages to a "wrapped" queue and cleanup:
        request = StoreRecords("metrics", wrapped_metrics, wrapped=True)
        metrics_stored = self.redis.ask(request)
        # Cleanup:
        if metrics_stored:
            cleaned_up = self.redis.ask(DeleteRecords("metrics", keys, wrapped=False))
        else:
            logger.warning(
                "[%s] Could not store %s Wrapped Metrics to a storage. "
                "Raw metrics are not cleaned.",
                self.name,
                len(wrapped_metrics),
            )
            cleaned_up = False
        if metrics_stored and not cleaned_up:
            logger.error(
                "[%s] Wrapped metrics were stored, but %s raw metrics "
                "were not cleaned up. Metrics can be duplicated!",
                self.name,
                len(b_records),
            )
        return metrics_stored and cleaned_up


class MetricsMerger:
    """
    MetricsMerger class is a part of a MergeAggregator workflow.

    It's responsible for separating raw metrics into groups and creating
    MergedMetric objects from raw data.
    Then these MergedMetric objects can be processed by a MetricsWrapper.
    """

    @staticmethod
    def group_metric_keys(
        metric_keys: List[Tuple[bytes, float]], interval: int
    ) -> List[List[bytes]]:
        """
        Takes a list of raw keys information, received from a storage and
        produces a list of groups of keys of metrics emitted during the
        same `interval` seconds.

        Input example:
        [(b"key-1", 9), (b"key-2", 11), (b"key-3", 18), (b"key-4", 21)]
        With default interval that is 10 output will be:
        [[b"key-1"], [b"key-2", b"key-3"], [b"key-4"]]
        They represent groups of events, happened between:
        Seconds: [[0-9], [10-19], [20-29]]

        Args:
            metric_keys: List of tuples (key, timestamp).
            interval: Length of a `flush interval` for aggregation.
        Returns: List of lists with grouped metrics keys.
        """
        grouped = groupby(metric_keys, lambda record: record[1] // interval)
        result = [[keys for keys, t_stamps in group[1]] for group in grouped]
        return result

    @classmethod
    def merge_metrics(cls, b_records: List[bytes], interval: int) -> List[MergedMetric]:
        """
        Takes a list of bytes presumably representing JSON encoded raw
        metrics and tries to cast them to a list of MergedMetrics.

        1. List of bytes is casted into MergedMetric objects.
        2. MergedMetrics are being grouped by their unique type id.
        3. Grouped MergedMetrics are merged into a list of single MergedMetrics.

        Args:
            b_records: List of bytes objects representing raw metrics.
            interval: Flush interval value.
        Returns: List of MergedMetric objects.
        """
        single_metrics = filter(
            None, [cls._cast_to_metric(record, interval) for record in b_records]
        )
        grouped_metrics = groupby(single_metrics, lambda metric: metric.id)
        merged_metrics = [
            reduce(lambda a, b: a + b, metrics) for _, metrics in grouped_metrics
        ]
        return merged_metrics

    @staticmethod
    def _cast_to_metric(b_record: bytes, interval: int) -> Optional[MergedMetric]:
        """
        Gets a bytes object and tries to decode to a JSON it and cast it
        to a MergedMetric object.

        If it can't be casted for some reason, returns None.

        Args:
            b_record: Bytes object, presumably representing a raw metric.
            interval: Flush interval value.
        Return: MergedMetric object or None.
        """
        try:
            dict_metric = json.loads(b_record)
            merged_metric = MergedMetric(
                metric=dict_metric["metric"],
                type=dict_metric["type"],
                timestamps=[dict_metric["timestamp"]],
                values=[dict_metric["value"]],
                tags=dict_metric.get("tags"),
                interval=interval,
            )
        except (json.JSONDecodeError, TypeError, KeyError):
            return None
        return merged_metric



"""
Chouette __main__ entry point.
"""
import logging
import sys
import time
from typing import Any, List, Type

from pythonjsonlogger import jsonlogger  # type: ignore

from chouette_iot import Scheduler, ChouetteConfig, Cancellable
from chouette_iot._singleton_actor import SingletonActor
from chouette_iot.metrics import MetricsCollector, MetricsAggregator, MetricsSender

logger = logging.getLogger("chouette-iot")


class Chouette:
    """
    Chouette entry point class that creates vital actors and their timers.
    """

    @staticmethod
    def setup_logging(log_level: str) -> None:
        """
        Configures logger to use JSON format and set s the desired log level.

        Args:
            log_level: Desired log level.
        Returns: None
        """
        stdout_handler = logging.StreamHandler(sys.stdout)
        json_formatter = jsonlogger.JsonFormatter(
            fmt="%(asctime)s %(levelname)s %(message)s", datefmt="%Y-%m-%dT%H:%M:%S%z"
        )
        stdout_handler.setFormatter(json_formatter)
        logging.basicConfig(level=log_level, handlers=[stdout_handler])

    @staticmethod
    def schedule_call(
        interval: float, actor_class: Type[SingletonActor], message: Any
    ) -> Cancellable:
        """
        Uses chouette-iot.Scheduler to periodically send a message to an actor of
        a specified class at some fixed rate.

        Args:
            interval: How often a message must be sent to an actor.
            actor_class: Class of a SingleActor child class to start.
            message: Message to send.
        Returns: Cancellable object.
        """
        actor_ref = actor_class.get_instance()
        initial_delay = interval - (time.time() % interval)
        timer = Scheduler.schedule_at_fixed_rate(
            initial_delay, interval, actor_ref.tell, message
        )
        return timer

    @classmethod
    def run(cls) -> List[Cancellable]:
        """
        Reads configuration from environment variables and creates timers to
        send messages to created actors.

        It starts a Sender actor and an Aggregator actor.
        If COLLECTOR_PLUGINS environment variable is set, it also starts a
        Collector plugin.

        Returns: List of Cancellables.
        """
        timers = []
        config = ChouetteConfig()
        cls.setup_logging(config.log_level)
        # Sender actor:
        timers.append(cls.schedule_call(config.release_interval, MetricsSender, "send"))
        # Aggregator actor:
        timers.append(
            cls.schedule_call(config.aggregate_interval, MetricsAggregator, "aggregate")
        )
        # Collector actor:
        if not config.collector_plugins:
            return timers
        timers.append(
            cls.schedule_call(config.capture_interval, MetricsCollector, "collect")
        )
        return timers


if __name__ == "__main__":
    logger.info("Starting Chouette-Iot.")
    Chouette.run()



"""
chouette.metrics module
"""
from ._metrics import MergedMetric, RawMetric, WrappedMetric
from ._aggregator import MetricsAggregator
from ._collector import MetricsCollector
from ._sender import MetricsSender

__all__ = [
    "MergedMetric",
    "MetricsAggregator",
    "MetricsCollector",
    "MetricsSender",
    "RawMetric",
    "WrappedMetric",
]



"""
MetricsWrapper abstract class.
"""
# pylint: disable=too-few-public-methods
from abc import ABC, abstractmethod
from typing import List

from chouette_iot.metrics import MergedMetric, WrappedMetric

__all__ = ["MetricsWrapper"]


class MetricsWrapper(ABC):
    """
    MetricsWrapper is the very heart of the Aggregation system.

    Its purpose is to take "raw" metrics sent by other applications and to
    cast them to standard Datadog metrics in a way that you prefer.
    """

    @classmethod
    def wrap_metrics(cls, merged_metrics: List[MergedMetric]) -> List[WrappedMetric]:
        """
        This is the only public method of a MetricsWrapper.

        It should take merged "raw" metrics and provide a list of wrapped
        metrics ready to be sent to Datadog.

        Args:
            merged_metrics: List of MergedMetric objects with raw metrics.
        Returns: List of WrappedMetric objects ready to be sent to Datadog.
        """
        metrics = [cls._wrap_metric(metric) for metric in merged_metrics]
        return sum(metrics, [])

    @classmethod
    @abstractmethod
    def _wrap_metric(cls, merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        Method that takes a single Merged Metric and casts it to a list
        of Wrapped Metrics.

        Args:
            merged_metric: A single MergedMetric to process.
        Returns: List of produced WrappedMetric objects.
        """
        raise NotImplementedError(
            "Use concrete Wrapper implementation."
        )  # pragma: no cover



"""
chouette.storages.messages module

This file describes objects that other actors should use as messages to
communicate to Storage actors.

These messages are being sent from clients to a Storage.
Both `ask` and `tell` patterns can be used to send them, but since storing
is a crucial part of any data processing, it's heavily recommended to use
`ask` pattern while communicating to Storages.
"""
# pylint: disable=too-few-public-methods
from typing import Iterable, List

__all__ = [
    "CleanupOutdatedRecords",
    "CollectKeys",
    "CollectValues",
    "DeleteRecords",
    "StoreRecords",
]


class CleanupOutdatedRecords:
    """
    Datadog rejects metrics older than 4 hours so there is no sense in
    trying to dispatch them.

    This message is being sent by a Sender actor to cleanup too old
    metrics and to let Sender dispatch only valid metrics.

    It cleans only ready to dispatch ("wrapped") records, so there is
    no "wrapped" flag that all other messages have.
    """

    __slots__ = ["data_type", "ttl", "wrapped"]

    def __init__(self, data_type: str, wrapped: bool, ttl: int = 14400):
        """
        Args:
            data_type: Type of data to cleanup. E.g: 'metrics'.
            ttl: Maximum metric lifetime in seconds.
        """
        self.data_type = data_type
        self.ttl = ttl
        self.wrapped = wrapped

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            f"<{self.__class__.__name__}:{self.data_type}:"
            f"wrapped={self.wrapped}:ttl={self.ttl}>"
        )


class CollectKeys:
    """
    This message initiates collection of record keys from a queue.

    Keys are being returned as a list of Tuples: (key: bytes, timestamp: int).
    """

    __slots__ = ["data_type", "wrapped", "amount", "reversed"]

    def __init__(self, data_type: str, wrapped: bool, amount: int = 0):
        """
        Args:
            data_type: Type of data to collect. E.g.: 'metrics'.
            wrapped: Whether a Storage should collect from a queue of
                     processed data.
            amount: Maximum number of keys that we want to collect. 0 is all.
        """
        self.amount = amount
        self.data_type = data_type
        self.wrapped = wrapped

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            f"<{self.__class__.__name__}:{self.data_type}:"
            f"wrapped={self.wrapped}:amount={self.amount}>"
        )


class CollectValues:
    """
    This message initiates collection of record values from a queue.

    Values are being returned as a list of bytes.
    """

    __slots__ = ["data_type", "keys", "wrapped"]

    def __init__(self, data_type: str, keys: List[bytes], wrapped: bool):
        """
        Args:
            data_type: Type of data to collect. E.g.: 'metrics'.
            wrapped: Whether a Storage should collect from a queue of
                     processed data.
            keys: List of keys of the records that we want to collect.
        """
        self.data_type = data_type
        self.keys = keys
        self.wrapped = wrapped

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            f"<{self.__class__.__name__}:{self.data_type}:"
            f"wrapped={self.wrapped}:keys_number={len(self.keys)}>"
        )


class DeleteRecords:
    """
    This message initiates deletion of record values from a queue.
    """

    __slots__ = ["data_type", "keys", "wrapped"]

    def __init__(self, data_type: str, keys: List[bytes], wrapped: bool):
        """
        Args:
            data_type: Type of data to delete. E.g.: 'metrics'.
            wrapped: Whether a Storage should delete from a queue of
                     processed data.
            keys: List of keys of the records that we want to delete.
        """
        self.data_type = data_type
        self.keys = keys
        self.wrapped = wrapped

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            f"<{self.__class__.__name__}:{self.data_type}:"
            f"wrapped={self.wrapped}:keys_number={len(self.keys)}>"
        )


class StoreRecords:
    """
    This message initiates deletion of record values in a queue.
    """

    __slots__ = ["data_type", "records", "wrapped"]

    def __init__(self, data_type: str, records: Iterable, wrapped: bool):
        """
        Args:
            data_type: Type of data to store. E.g.: 'metrics'.
            wrapped: Whether a Storage should store to a queue of
                     processed data.
            records: Iterable of preprocessed objects with `asdict` method.
                     E.g.: WrappedMetric.
        """
        self.data_type = data_type
        self.records = list(records)
        self.wrapped = wrapped

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return (
            f"<{self.__class__.__name__}:{self.data_type}:"
            f"wrapped={self.wrapped}:records_number={len(self.records)}>"
        )



"""
Actor that handles all interactions with the Redis storage.
"""
# pylint: disable=too-few-public-methods
import json
import logging
import time
from typing import Any, List, Tuple, Union
from uuid import uuid4

from pydantic import BaseSettings  # type: ignore
from redis import Redis, RedisError

from chouette_iot._singleton_actor import SingletonActor
from ._redis_messages import GetRedisQueues, GetHashSizes
from .messages import (
    CleanupOutdatedRecords,
    CollectKeys,
    CollectValues,
    DeleteRecords,
    StoreRecords,
)

__all__ = ["RedisStorage"]

logger = logging.getLogger("chouette-iot")


class RedisConfig(BaseSettings):
    """
    RedisStorage environment configuration object.
    Reads Redis' host and port from environment variables if called.
    """

    redis_host: str = "redis"
    redis_port: int = 6379


class RedisStorage(SingletonActor):
    """
    Redis handling Singleton Actor.

    Handles requests to collect, store, and delete different kinds of records
    from the Redis storage.

    Records themselves are stored in Redis hashes like this: {key: metric body}
    Their keys are being stored in Sorted Sets sorted by timestamps.

    Records can be `raw` (unprocessed, received from clients) or `wrapped`
    (prepared to dispatch). They are being stored in different queues.

    The queue for raw metrics is named like this:
    Sorted set: `chouette:raw:metrics.keys`.
    Hash: `chouette:raw:metrics.values`.

    This pattern `chouette:(raw/wrapped):(record type).(keys/values)` is used
    for all the queues.

    Is intentionally expected to be almost always used with `ask` pattern
    to ensure that consumers always execute their logic in a correct order.
    """

    def __init__(self):
        """
        RedisStorage uses RedisConfig to create a Redis object.
        """
        super().__init__()
        config = RedisConfig()
        self.redis = Redis(host=config.redis_host, port=config.redis_port)
        # Different versions of Redis use different HSET command formats:
        redis_version = self.redis.info().get("redis_version")
        self.redis_version = int(redis_version.split(".")[0])

    def on_receive(self, message: Any) -> Union[list, bool, None]:
        """
        Messages handling routine.

        It takes any message, but it will actively process only valid
        storage messages from the `chouette.storages.messages` package.

        All other messages do nothing and receive None if `ask` was used
        to send them.

        Args:
            message: Anything. Expected to be a valid storage message.
        Returns: Either a List of bytes or a bool.
        """
        logger.debug("[%s] Received %s.", self.name, message)
        if isinstance(message, CleanupOutdatedRecords):
            return self._cleanup_outdated(message)

        if isinstance(message, CollectKeys):
            return self._collect_keys(message)

        if isinstance(message, CollectValues):
            return self._collect_values(message)

        if isinstance(message, DeleteRecords):
            return self._delete_records(message)

        if isinstance(message, GetHashSizes):
            return self._get_hash_sizes(message)

        if isinstance(message, GetRedisQueues):
            return self._get_redis_queues(message)

        if isinstance(message, StoreRecords):
            return self._store_records(message)

        return None

    def _cleanup_outdated(self, request: CleanupOutdatedRecords) -> bool:
        """
        Cleans up outdated records in a specified queue.

        Datadog rejects metrics older than 4 hours (default TTL), so before
        trying to dispatch anything Chouette cleans up outdated metrics.

        Args:
            request: CleanupOutdated message with record type and TTL.
        Returns: Boolean that says whether execution was successful.
        """
        queue_name, set_name, hash_name = self._get_queue_names(request)
        threshold = time.time() - request.ttl
        try:
            outdated_keys = self.redis.zrangebyscore(set_name, 0, threshold)
            if not outdated_keys:
                logger.debug(
                    "[%s] No outdated records to cleanup in a queue '%s'",
                    self.name,
                    queue_name,
                )
                return True
            pipeline = self.redis.pipeline()
            pipeline.zremrangebyscore(set_name, 0, threshold)
            pipeline.hdel(hash_name, *outdated_keys)
            pipeline.execute()
            logger.debug(
                "[%s] Cleaned %s outdated records from a queue '%s'.",
                self.name,
                len(outdated_keys),
                queue_name,
            )
        except RedisError as error:
            logger.warning(
                "[%s] Could not cleanup records in a queue '%s' due to: '%s'.",
                self.name,
                queue_name,
                error,
            )
            return False
        return True

    def _collect_keys(self, request: CollectKeys) -> List[Tuple[bytes, int]]:
        """
        Tries to collect keys from a specified queue.

        CollectKeys message has the following properties:
        * data_type - type of a queue, e.g.: 'metrics'.
        * wrapped - whether that's a queue of processed records or not.
        * amount - how many keys should be collected. 0 means `all of them`.

        It returns a list of tuples with keys and their timestamps:
        (key: bytes, timestamp: int).

        Args:
            request: CollectKeys message.
        Returns: List of collected keys as tuples.
        """
        queue_name, set_name, _ = self._get_queue_names(request)
        try:
            keys = self.redis.zrange(set_name, 0, request.amount - 1, withscores=True)
        except RedisError as error:
            logger.warning(
                "[%s] Could not collect keys from a queue '%s' due to: '%s'.",
                self.name,
                queue_name,
                error,
            )
            return []
        logger.debug(
            "[%s] Collected %s keys from a queue '%s'.",
            self.name,
            len(keys),
            queue_name,
        )
        return keys

    def _collect_values(self, request: CollectValues) -> List[bytes]:
        """
        Tries to collect values by keys from a specified queue.

        Receives a CollectValues message that contains a list of keys.
        It goes to this queue's hash and gets corresponding records as bytes.

        Args:
            request: CollectValues message with specified keys.
        Returns: List of collected values.
        """
        queue_name, _, hash_name = self._get_queue_names(request)
        if not request.keys:
            logger.debug(
                "[%s] No keys were specified to collect values for a queue '%s'.",
                self.name,
                queue_name,
            )
            return []
        try:
            raw_values = self.redis.hmget(hash_name, *request.keys)
            values: List[bytes] = list(filter(None, raw_values))
        except RedisError as error:
            logger.warning(
                "[%s] Could not collect records from a queue '%s' due to: '%s'.",
                self.name,
                queue_name,
                error,
            )
            return []
        logger.debug(
            "[%s] Collected %s records from a queue '%s'.",
            self.name,
            len(values),
            queue_name,
        )
        return values

    def _delete_records(self, request: DeleteRecords) -> bool:
        """
        Tries to delete records with specified keys.

        Args:
            request: DeleteRecords message with specified keys.
        Returns: Boolean that says whether execution was successful.
        """
        queue_name, set_name, hash_name = self._get_queue_names(request)
        if not request.keys:
            logger.debug(
                "[%s] Nothing to delete from a queue '%s'.", self.name, queue_name
            )
            return True
        pipeline = self.redis.pipeline()
        pipeline.zrem(set_name, *request.keys)
        pipeline.hdel(hash_name, *request.keys)
        try:
            pipeline.execute()
        except RedisError as error:
            logger.warning(
                "[%s] Could not remove %s records from a queue '%s' due to: '%s'.",
                self.name,
                len(request.keys),
                queue_name,
                error,
            )
            return False
        logger.debug(
            "[%s] Deleted %s records from a queue '%s'.",
            self.name,
            len(request.keys),
            queue_name,
        )
        return True

    def _get_hash_sizes(self, request: GetHashSizes) -> List[Tuple[str, int]]:
        """
        Returns a list of tuples with hashes names and sizes.

        Args:
            request: GetHashSizes message with a list of hashes.
        Return: List of tuples with hashes names and sizes.
        """
        try:
            hash_sizes = [
                (
                    hash_name.decode() if isinstance(hash_name, bytes) else hash_name,
                    int(self.redis.hlen(hash_name)),
                )
                for hash_name in request.hashes
            ]
        except RedisError as error:
            logger.warning(
                "[%s] Could not calculate hash sizes due to: '%s'.", self.name, error
            )
            return []
        return hash_sizes

    def _get_redis_queues(self, request: GetRedisQueues) -> List[bytes]:
        """
        Gets a list of Redis Keys (sets, hashes, lists, etc) using a specified
        key name pattern.

        Args:
            request: GetRedisKeys message with a specified pattern.
        Returns: List of found Redis Keys.
        """
        try:
            redis_keys = self.redis.keys(request.pattern)
        except RedisError as error:
            logger.warning(
                "[%s] Could not collect Redis keys for a pattern %s due to: '%s'.",
                self.name,
                request.pattern,
                error,
            )
            return []
        return redis_keys

    def _store_records(self, request: StoreRecords) -> bool:
        """
        Tries to store received records to a queue.

        It automatically generates a unique id for every record and stores
        its content under this id both to a set and a hash.

        If it can't cast one of the records to a dict via `asdict()` method,
        it ignores this record and tries to store all other records.

        Args:
            request: StoreRecords with an iterable of suitable objects.
        Returns: Boolean that says whether execution was successful.
        """
        queue_name, set_name, hash_name = self._get_queue_names(request)
        pipeline = self.redis.pipeline()
        records_list = list(request.records)
        keys = {}
        values = {}
        for record in records_list:
            try:
                record_value = json.dumps(record.asdict())
            except AttributeError:
                continue
            record_key = str(uuid4())
            keys[record_key] = record.timestamp
            values[record_key] = record_value
        stored_metrics = len(values)
        if not values:
            logger.debug(
                "[%s] Nothing to store to a queue '%s'.", self.name, queue_name
            )
            return True
        try:
            pipeline.zadd(set_name, mapping=keys)
            if self.redis_version >= 4:
                # From Redis 4.0.0 HMSET command is deprecated.
                pipeline.hset(hash_name, mapping=values)
            else:
                # Before Redis 4.0.0 HSET command took only 2 arguments:
                pipeline.hmset(hash_name, mapping=values)
            pipeline.execute()
        except (RedisError, TypeError) as error:
            logger.warning(
                "[%s] Could not store %s/%s records to queue '%s' due to: '%s'.",
                self.name,
                stored_metrics,
                len(records_list),
                queue_name,
                error,
            )
            return False
        logger.debug(
            "[%s] Stored %s/%s records to a queue '%s'.",
            self.name,
            stored_metrics,
            len(records_list),
            queue_name,
        )
        return True

    @staticmethod
    def _get_queue_names(request: Any) -> Tuple[str, str, str]:
        """
        Generates queue, set and hash name for a queue depending on a request.

        Args:
            request: One of `chouette.storage.messages` objects.
        Return: Tuple of a queue name, a set name and a hash name as strings.
        """
        queue_type = "wrapped" if request.wrapped else "raw"
        queue_name = f"chouette:{request.data_type}:{queue_type}"
        set_name = f"{queue_name}.keys"
        hash_name = f"{queue_name}.values"
        return queue_name, set_name, hash_name



"""
CollectorPlugin: Abstract class for all metric collectors.
"""
# pylint: disable=too-few-public-methods
from abc import ABC
from typing import Iterator, List, Optional, Tuple

from chouette_iot.metrics import WrappedMetric

__all__ = ["CollectorPlugin"]


class CollectorPlugin(ABC):
    """
    Abstract class for all metric collectors.
    """

    @staticmethod
    def _wrap_metrics(
        metrics: List[Tuple[str, float]],
        timestamp: float = None,
        tags: Optional[List[str]] = None,
        metric_type: str = "gauge",
    ) -> Iterator[WrappedMetric]:
        """
        Generates a list of WrappedMetric objects.

        Args:
            metrics: List of (metric_name, metric_value) tuples.
            timestamp: Metric collection timestamp.
            tags: Metric tags.
            metric_type: Metric type.
        Returns: Iterator over WrappedMetric objects.
        """
        wrapped_metrics: List[WrappedMetric] = [
            WrappedMetric(
                metric=metric_name,
                type=metric_type,
                value=metric_value,
                timestamp=timestamp,
                tags=tags,
            )
            for metric_name, metric_value in metrics
            if metric_value
        ]
        return iter(wrapped_metrics)



"""
Concrete simplified implementation of a Datadog wrapper.
"""
# pylint: disable=too-few-public-methods
import math
from typing import Any, List, Set

from pydantic import BaseSettings  # type: ignore

from chouette_iot.metrics import MergedMetric, WrappedMetric
from ._metrics_wrapper import MetricsWrapper

__all__ = ["DatadogWrapper"]


class DatadogWrapperConfig(BaseSettings):
    """
    Optional Wrapper configuration object.
    """

    histogram_aggregates: List[str] = ["max", "median", "avg", "count"]
    histogram_percentiles: List[float] = [0.95]


class DatadogWrapper(MetricsWrapper):
    """
    Datadog wrapper tries to implement the same behaviour as Datadog agent
    that is described here:
    https://docs.datadoghq.com/developers/metrics/types/

    That is not some really accurate implementation of this behaviour, that's
    an attempt to have a Datadog-like aggregation strategy.

    Supported metrics:
    COUNT - sends a sum of values received during a flush interval.
    GAUGE - sends the last value received during a flush interval.
    RATE - sends a number of events happened during 1 seconds of a
           flush interval.
    SET - sends a count of unique elements send in a metric during a
           flush interval.
    HISTOGRAM - sends a set of different metrics according to
           histogram_aggregates and histogram_percentiles configuration.

    DISTRIBUTION metric type is NOT supported.
    """

    histogram_percentiles = DatadogWrapperConfig().histogram_percentiles
    histogram_aggregates = DatadogWrapperConfig().histogram_aggregates

    @classmethod
    def _wrap_metric(cls, merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        Facade wrapper function of the Datadog Wrapper.

        It processes: COUNT, GAUGE, RATE, SET and HISTOGRAM metrics.

        All other kinds of metrics are ignored.

        Args:
            merged_metric: MergedMetric to wrap.
        Returns: List of WrappedMetric produced by the wrapping method.
        """
        methods = {
            "count": cls._wrap_count,
            "rate": cls._wrap_rate,
            "gauge": cls._wrap_gauge,
            "set": cls._wrap_set,
            "histogram": cls._wrap_histogram,
        }
        method = methods.get(merged_metric.type)
        if not method:
            return []
        return method(merged_metric)

    @staticmethod
    def _wrap_count(merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        Count metric is being simply wrapped into a metric whose value is a
        sum of all the values and whose timestamp is the earliest timestamp
        in a sequence.

        Args:
            merged_metric: MergedMetric to wrap.
        Returns: List of WrappedMetric produced by the wrapping method.
        """
        count_metric = WrappedMetric(
            metric=merged_metric.metric,
            type=merged_metric.type,
            timestamp=min(merged_metric.timestamps),
            value=sum(merged_metric.values),
            tags=merged_metric.s_tags,
            interval=merged_metric.interval,
        )
        return [count_metric]

    @classmethod
    def _wrap_rate(cls, merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        Rate metric represents an approximate rate of event occurrences per
        one second of the flush interval.
        Like count it calculates the sum of all the values but then this
        value is divided by a number of seconds in the flush interval.
        Timestamp is the earliest timestamp in a sequence.

        Args:
            merged_metric: MergedMetric to wrap.
        Returns: List of WrappedMetric produced by the wrapping method.
        """
        flush_interval = float(merged_metric.interval)
        rate_metric = WrappedMetric(
            metric=merged_metric.metric,
            type=merged_metric.type,
            timestamp=min(merged_metric.timestamps),
            value=sum(merged_metric.values) / flush_interval,
            tags=merged_metric.s_tags,
            interval=merged_metric.interval,
        )
        return [rate_metric]

    @staticmethod
    def _wrap_gauge(merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        Gauge metric sends the latest received value without additional
        calculations. Its timestamp however is the earliest timestamp
        in a sequence.

        Args:
            merged_metric: MergedMetric to wrap.
        Returns: List of WrappedMetric produced by the wrapping method.
        """
        data_pairs = zip(merged_metric.values, merged_metric.timestamps)
        value, _ = max(data_pairs, key=lambda pair: pair[1])
        gauge_metric = WrappedMetric(
            metric=merged_metric.metric,
            type=merged_metric.type,
            timestamp=min(merged_metric.timestamps),
            value=value,
            tags=merged_metric.s_tags,
        )
        return [gauge_metric]

    @staticmethod
    def _wrap_set(merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        Set metric is expected to have `Lists` in values. There lists
        should represent sets of data. Wrapper takes these lists and
        collects how many unique elements they contained.

        E.g.:
        On second 1 metric 'users' was produced with the following list:
        [Alice, Bob].
        On second 9 metric 'users' was produces with the following list:
        [Bob, Carol].

        This wrapper takes both lists, merge them into a single list:
        [Alice, Bob, Bob, Carol]
        and then calculates the number of unique elements.
        In this case it's 3.

        Args:
            merged_metric: MergedMetric to wrap.
        Returns: List of WrappedMetric produced by the wrapping method.
        """
        try:
            values_set: Set[Any] = set(sum(merged_metric.values, []))
            set_count_metric = WrappedMetric(
                metric=merged_metric.metric,
                type="count",
                timestamp=min(merged_metric.timestamps),
                value=len(values_set),
                tags=merged_metric.s_tags,
                interval=merged_metric.interval,
            )
            return [set_count_metric]
        except TypeError:
            return []

    @classmethod
    def _wrap_histogram(cls, merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        Histogram metric wrapper implementation.

        It depends on HISTOGRAM_PERCENTILES and HISTOGRAM_AGGREGATES
        configuration.

        These configuration options are described here:
        https://docs.datadoghq.com/developers/metrics/types/#?tab=histogram

        By default it generates exactly the same 5 metrics:
        `avg`, `count`, `max`, `median`, `95percentile`
        Other metrics like `sum`, `min` or other percentiles can be
        configured via configuration mentioned above.

        To avoid bringing Numpy just to calculate percentiles, a `_percentile`
        method is used. In tests it provided the same values as numpy.

        Args:
            merged_metric: MergedMetric to wrap.
        Returns: List of WrappedMetric produced by the wrapping method.
        """
        interval = float(merged_metric.interval)
        timestamp = min(merged_metric.timestamps)
        tags = merged_metric.tags
        values = merged_metric.values
        name = merged_metric.metric
        metrics_count = len(values)
        metrics_to_generate = [
            (f"{name}.avg", "gauge", sum(values) / metrics_count, None),
            (f"{name}.count", "rate", metrics_count / interval, int(interval)),
            (f"{name}.sum", "gauge", sum(values), None),
            (f"{name}.min", "gauge", min(values), None),
            (f"{name}.max", "gauge", max(values), None),
            (f"{name}.median", "gauge", cls._percentile(values, 0.5), None),
        ]
        percentiles_metrics_to_generate = [
            (
                f"{name}.{int(percentile * 100)}percentile",
                "gauge",
                cls._percentile(values, percentile),
                None,
            )
            for percentile in cls.histogram_percentiles
        ]
        metrics_to_generate.extend(percentiles_metrics_to_generate)
        generated_metrics = [
            WrappedMetric(
                metric=metric_name,
                type=metric_type,
                timestamp=timestamp,
                value=value,
                tags=tags,
                interval=interval,
            )
            for metric_name, metric_type, value, interval in metrics_to_generate
            if "percentile" in metric_name
            or metric_name.split(".")[-1] in cls.histogram_aggregates
        ]
        return generated_metrics

    @staticmethod
    def _percentile(data_set: List[float], percent: float) -> float:
        """
        Since we just need to calculate percentile and median and we don't
        want to bring the whole numpy here for this task.

        Args:
            data_set: List of float or integer values.
            percent: What percentile should be returned.
        return: Float value of a percentile.
        """
        sorted_ds = sorted(data_set)
        idx = (len(sorted_ds) - 1) * percent
        if idx % 1 == 0:
            return sorted_ds[int(idx)]
        upper_idx = math.ceil(idx)
        lower_idx = math.floor(idx)
        right_value = sorted_ds[upper_idx] * (idx - lower_idx)
        left_value = sorted_ds[lower_idx] * (upper_idx - idx)
        return left_value + right_value



"""
Concrete implementation of a simple wrapper.
"""
# pylint: disable=too-few-public-methods
from typing import List

from ._metrics_wrapper import MetricsWrapper

__all__ = ["SimpleWrapper"]

from .. import WrappedMetric, MergedMetric


class SimpleWrapper(MetricsWrapper):
    """
    SimpleWrapper class takes only two kinds of metrics:
    `count` and `gauge`. Any metrics that is not `count` is being cast
    into a `gauge` metrics that, in fact, is not a `gauge` in Datadog sense.

    Unlike standard Datadog, it doesn't send the last value for the `gauge`
    metric. Instead it sends an average value along with a count of metrics
    used to calculate this average value.

    Received metric types:
    `count`:
            sends a single `count` metric with a sum of values.
     anything else:
            sends a `gauge` metric with an average value and a `.count`
            metric with a number of metrics used to calculate the avg.
    """

    @classmethod
    def _wrap_metric(cls, merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        This method determines what method to call to wrap this exact metric
        and calls it.

        For this wrapper it knows only two methods - a method to wrap `count`
        metrics and a method to wrap anything else.

        Args:
            merged_metric: MergedMetric to wrap.
        Returns: List of WrappedMetric produced by the wrapping method.
        """
        methods = {"count": cls._wrap_count}
        metric_type = merged_metric.type
        return methods.get(metric_type, cls._wrap_average)(merged_metric)

    @staticmethod
    def _wrap_count(merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        Count metric is being simply wrapped into a metric whose value is a
        sum of all the values and whose timestamp is the latest timestamp
        in a sequence.

        Args:
            merged_metric: MergedMetric to wrap.
        Returns: List of WrappedMetric produced by the wrapping method.
        """
        count_metric = WrappedMetric(
            metric=merged_metric.metric,
            type=merged_metric.type,
            timestamp=max(merged_metric.timestamps),
            value=sum(merged_metric.values),
            tags=merged_metric.s_tags,
            interval=merged_metric.interval,
        )
        return [count_metric]

    @staticmethod
    def _wrap_average(merged_metric: MergedMetric) -> List[WrappedMetric]:
        """
        Average metric is represented by two WrappedMetrics:
        1. `gauge` type metric whose value is the average of the original
        `values` list.
        2. `count` metric whose value is the number of elements in the
        `values` list used for average calculation. Its name is basically
        the name of the `gauge` metrics with '.count' suffix.

        Args:
            merged_metric: MergedMetric to wrap.
        Returns: List of WrappedMetric produced by the wrapping method.
        """
        metrics_count = len(merged_metric.values)
        average = sum(merged_metric.values) / metrics_count
        timestamp = max(merged_metric.timestamps)
        average_metric = WrappedMetric(
            metric=merged_metric.metric,
            type="gauge",
            timestamp=timestamp,
            value=average,
            tags=merged_metric.s_tags,
        )
        count_metric = WrappedMetric(
            metric=f"{merged_metric.metric}.count",
            type="count",
            timestamp=timestamp,
            value=metrics_count,
            tags=merged_metric.s_tags,
            interval=merged_metric.interval,
        )
        return [average_metric, count_metric]



"""
Redis specific messages, being used to collect stats about Redis queues.
"""
# pylint: disable=too-few-public-methods

from typing import Iterable, Union


class GetHashSizes:
    """
    THis message initiates collection of Redis Hashes sizes.
    """

    __slots__ = ["hashes"]

    def __init__(self, hashes: Iterable[Union[bytes, str]]):
        self.hashes = hashes


class GetRedisQueues:
    """
    This message initiates collection of Redis entities names.
    """

    __slots__ = ["pattern"]

    def __init__(self, pattern: str):
        self.pattern = pattern



"""
Main Chouette module definition.
"""
from chouette_iot._configuration import ChouetteConfig
from chouette_iot._scheduler import Cancellable, Scheduler

__all__ = ["Cancellable", "Scheduler", "ChouetteConfig"]



"""
chouette.metrics.plugins
"""
# pylint: disable=too-few-public-methods
from typing import Dict, Optional, Type

from pykka import ActorRef  # type: ignore

from chouette_iot._singleton_actor import SingletonActor
from ._docker_collector import DockerCollector
from ._dramatiq_collector import DramatiqCollector
from ._host_collector import HostStatsCollector
from ._k8s_collector import K8sCollector
from ._tegrastats_collector import TegrastatsCollector

__all__ = [
    "DockerCollector",
    "DramatiqCollector",
    "HostStatsCollector",
    "K8sCollector",
    "PluginsFactory",
    "TegrastatsCollector",
]


class PluginsFactory:
    """
    PluginsFactory class creates plugins actors and returns their ActorRefs.
    """

    plugins: Dict[str, Type[SingletonActor]] = {
        "dramatiq": DramatiqCollector,
        "host": HostStatsCollector,
        "k8s": K8sCollector,
        "tegrastats": TegrastatsCollector,
        "docker": DockerCollector,
    }

    @classmethod
    def get_plugin(cls, plugin_name: str) -> Optional[ActorRef]:
        """
        Takes a plugin name and returns an ActorRef if such plugin exists.

        Args:
            plugin_name: Plugin name as a string.
        Returns: ActorRef or None.
        """
        plugin_class = cls.plugins.get(plugin_name)
        if not plugin_class:
            return None
        actor_ref: ActorRef = plugin_class.get_instance()
        return actor_ref



"""
chouette.metrics.plugins.DramatiqCollector
"""
# pylint: disable=too-few-public-methods
import logging
import re
from itertools import chain
from typing import Iterator, List, Tuple

from pykka import ActorDeadError  # type: ignore

from chouette_iot._singleton_actor import SingletonActor
from chouette_iot.metrics import WrappedMetric
from chouette_iot.storages import RedisStorage
from chouette_iot.storages._redis_messages import GetHashSizes, GetRedisQueues
from ._collector_plugin import CollectorPlugin
from .messages import StatsRequest, StatsResponse

__all__ = ["DramatiqCollector"]

logger = logging.getLogger("chouette-iot")


class DramatiqCollector(SingletonActor):
    """
    Actor that collects Dramatiq queues sizes.

    NB: Collectors MUST interact with plugins via `tell` pattern.
        `ask` pattern will return None.
    """

    def __init__(self):
        super().__init__()
        self.redis = RedisStorage.get_instance()

    def on_receive(self, message: StatsRequest) -> None:
        """
        On StatsRequest message collects Dramatiq queues sizes and
        sends them back in a StatsResponse message.

        On any other message does nothing.

        Args:
            message: Expected to be a StatsRequest message.
        """
        logger.debug("[%s] Received %s.", self.name, message)
        if isinstance(message, StatsRequest):
            hashes = self.redis.ask(GetRedisQueues("dramatiq:*.msgs"))
            sizes = self.redis.ask(GetHashSizes(hashes))
            stats = DramatiqCollectorPlugin.wrap_queues_sizes(sizes)
            if hasattr(message.sender, "tell"):
                try:
                    message.sender.tell(StatsResponse(self.name, stats))
                except ActorDeadError:
                    logger.warning(
                        "[%s] Requester is stopped. Dropping message.", self.name
                    )


class DramatiqCollectorPlugin(CollectorPlugin):
    """
    DramatiqPlugin that wraps received hashes sizes into WrappedMetrics.
    """

    @classmethod
    def wrap_queues_sizes(cls, sizes: List[Tuple[str, int]]) -> Iterator[WrappedMetric]:
        """
        Wraps received hashes sizes into WrappedMetrics.

        Args:
            sizes: List of tuples ("hash_name", hash_size as int).
        Returns: Iterator over WrappedMetric objects.
        """
        metrics = [
            cls._wrap_metrics(
                [("Chouette.dramatiq.queue_size", size)],
                tags=[f"queue:{re.sub(r'^dramatiq:|.msgs$', '', queue)}"],
            )
            for queue, size in sizes
        ]
        return chain.from_iterable(metrics)



"""
chouette.metrics.plugins.K8sCollector
"""
# pylint: disable=too-few-public-methods
import json
import logging
from itertools import chain
from typing import Any, Dict, Iterator, List, Optional, Tuple

import requests
from pydantic import BaseSettings, ValidationError  # type: ignore
from pykka import ActorDeadError  # type: ignore

from chouette_iot._singleton_actor import SingletonActor
from ._collector_plugin import CollectorPlugin
from .messages import StatsRequest, StatsResponse
from chouette_iot.metrics import WrappedMetric

__all__ = ["K8sCollector"]

logger = logging.getLogger("chouette-iot")


class K8sCollectorConfig(BaseSettings):
    """
    Environment variables based plugin configuration.

    K8S_STATS_SERVICE_IP is normally the IP of the node, it could be an
    external IP (e.g. 192.168.1.104) or an internal IP (e.g. 10.1.18.1). It's
    better to use an internal IP if your connectivity isn't too good. because
    if you rely on your external IP and connection disappears, the plugin will
    stop collecting metrics. In case of an internal IP that won't happen.

    K8S_STATS_SERVICE_PORT is a port of the Stats Service. By default it's
    10250.

    K8S_CERT_PATH is a path to a client certificate to pass the Stats Server
    authorization.
    In microk8s it's usually /var/snap/microk8s/current/certs/server.crt

    K8S_KEY_PATH is a path to a client key to pass the Stats Server
    authorization.
    In microk8s it's usually /var/snap/microk8s/current/certs/server.key

    K8S_METRICS is a structure that defines what metrics should be sent
    to Datasog.
    """

    k8s_stats_service_ip: str
    k8s_stats_service_port: int = 10250
    k8s_cert_path: str  # Path to server.crt for microk8s
    k8s_key_path: str  # Path to server.key for microk8s
    k8s_metrics: Dict[str, List[str]] = {"pods": ["memory", "cpu"], "node": ["inodes"]}


class K8sCollector(SingletonActor):
    """
    K8sCollector plugin collects stats about K8s node and pods via a K8s
    Stats Service.
    """

    def __init__(self):
        super().__init__()
        try:
            config = K8sCollectorConfig()
            self.k8s_url: str = f"https://{config.k8s_stats_service_ip}:{config.k8s_stats_service_port}/stats/summary"
            self.certs: Tuple[str, str] = (config.k8s_cert_path, config.k8s_key_path)
            self.k8s_metrics: Dict[str, List[str]] = config.k8s_metrics
        except ValidationError:
            self.k8s_url = None
            logger.warning(
                "[%s] Kubernetes configuration is not correct. Metrics won't be collected.",
                self.name,
                exc_info=True,
            )

    def on_receive(self, message: StatsRequest) -> None:
        """
        On StatsRequest message collects K8s pods stats and
        sends them back in a StatsResponse message.

        On any other message does nothing.

        Args:
            message: Expected to be a StatsRequest message.
        """
        logger.debug("[%s] Received %s.", self.name, message)
        if isinstance(message, StatsRequest) and self.k8s_url:
            stats = K8sCollectorPlugin.collect_stats(
                self.k8s_url, self.certs, self.k8s_metrics
            )
            if hasattr(message.sender, "tell"):
                try:
                    message.sender.tell(StatsResponse(self.name, stats))
                except ActorDeadError:
                    logger.warning(
                        "[%s] Requester is stopped. Dropping message.", self.name
                    )


class K8sCollectorPlugin(CollectorPlugin):
    """
    K8sCollectorPlugin handles collecting data from K8s and wrapping it
    into iterator of WrappedMetrics.
    """

    @classmethod
    def collect_stats(
        cls, url: str, certs: Tuple[str, str], to_collect: Dict[str, List[str]]
    ) -> Iterator[WrappedMetric]:
        """
        Gathers all the requested metrics and wraps them into an Iterator.

        If it couldn't get any data from K8s - returns an empty Iterator.

        Args:
            url: URL of a Stats Service to gather data from.
            certs: Tuple with paths of a client cert and a client key to
                   authorize.
            to_collect: Dict with metrics configuration.
        Returns: Iterator over WrappedMetric objects.
        """
        raw_metrics_dict = cls._get_raw_metrics(url, certs)
        if not raw_metrics_dict:
            return iter([])
        stats = [
            cls._parse_node_metrics(raw_metrics_dict, to_collect.get("node", [])),
            cls._parse_pods_metrics(raw_metrics_dict, to_collect.get("pods", [])),
        ]
        return chain.from_iterable(stats)

    @classmethod
    def _parse_node_metrics(
        cls, raw_metrics: Dict[str, Any], to_collect: List[str]
    ) -> Iterator[WrappedMetric]:
        """
        Gets a dict with a raw K8s Stats Service response and produces
        metrics based on its "node" content.

        This method doesn't return all the metrics that could be found
        in a Stats Service response.
        That's just a subset that can be extended with more fs data
        or networking data if necessary.

        Args:
            raw_metrics: Dict containing K8s Stats Service response.
            to_collect: List of node metrics to return.
        Returns: Iterator over WrappedMetric objects.
        """
        node = raw_metrics.get("node")
        if not node:
            return iter([])
        node_name = node.get("nodeName")
        tags = [f"node_name:{node_name}"]
        cpu = node.get("cpu", {}) if "cpu" in to_collect else {}
        ram = node.get("memory", {}) if "ram" in to_collect else {}
        network = node.get("network", {}) if "network" in to_collect else {}
        f_system = node.get("fs", {}) if "filesystem" in to_collect else {}
        inodes = (
            node.get("fs", {}).get("inodesFree") if "inodes" in to_collect else None
        )
        prefix = "Chouette.k8s.node"
        stats = [
            (f"{prefix}.cpu.usageNanoCores", cpu.get("usageNanoCores")),
            (f"{prefix}.cpu.usageCoreNanoSeconds", cpu.get("usageCoreNanoSeconds")),
            (f"{prefix}.memory.rssBytes", ram.get("rssBytes")),
            (f"{prefix}.memory.usageBytes", ram.get("usageBytes")),
            (f"{prefix}.memory.availableBytes", ram.get("availableBytes")),
            (f"{prefix}.memory.workingSetBytes", ram.get("workingSetBytes")),
            (f"{prefix}.network.rxBytes", network.get("rxBytes")),
            (f"{prefix}.network.txBytes", network.get("txBytes")),
            (f"{prefix}.fs.inodesFree", inodes),
            (f"{prefix}.fs.availableBytes", f_system.get("availableBytes")),
            (f"{prefix}.fs.usedBytes", f_system.get("usedBytes")),
        ]
        collected_metrics = [(name, value) for name, value in stats if value]
        return cls._wrap_metrics(collected_metrics, tags=tags)

    @classmethod
    def _parse_pods_metrics(
        cls, raw_metrics: Dict[str, Any], to_collect: List[str]
    ) -> Iterator[WrappedMetric]:
        """
        Just a wrapper for a _parse_pod_metrics method.

        Takes a raw_metrics dict, gets a list of pods from it and generate
        an Iterator of pods metrics using a _parse_pod_metrics method.

        Args:
            raw_metrics: Dict containing K8s Stats Service response.
            to_collect: List of pod metrics to return.
        Returns: Iterator over WrappedMetric objects.
        """
        pods: List[Dict[str, Any]] = raw_metrics.get("pods", [])
        metrics = [
            cls._parse_pod_metrics(pod_metrics, to_collect) for pod_metrics in pods
        ]
        return chain.from_iterable(metrics)

    @classmethod
    def _parse_pod_metrics(
        cls, pod: Dict[str, Any], to_collect: List[str]
    ) -> Iterator[WrappedMetric]:
        """
        Gets a dict with a pod stats description and casts it into an
        Iterator over metrics.

        That's just a tiny subset of metrics returned by K8s Stats Service,
        it can be expanded if necessary.

        Args:
            pod: Dict containing pod data from K8s Stats Service response.
            to_collect: List of pod metrics to return.
        Returns: Iterator over WrappedMetric objects.
        """
        pod_ref: Optional[Dict[str, str]] = pod.get("podRef")
        if not pod_ref:
            return iter([])
        tags = [
            f"namespace:{pod_ref.get('namespace')}",
            f"pod_name:{pod_ref.get('name')}",
        ]
        cpu = pod.get("cpu", {}) if "cpu" in to_collect else {}
        ram = pod.get("memory", {}) if "memory" in to_collect else {}
        network = pod.get("network", {}) if "network" in to_collect else {}
        prefix = "Chouette.k8s.pod"
        stats = [
            (f"{prefix}.cpu.usageNanoCores", cpu.get("usageNanoCores")),
            (f"{prefix}.memory.rssBytes", ram.get("rssBytes")),
            (f"{prefix}.memory.usageBytes", ram.get("usageBytes")),
            (f"{prefix}.network.rxBytes", network.get("rxBytes")),
            (f"{prefix}.network.txBytes", network.get("txBytes")),
        ]
        collected_metrics = [(name, value) for name, value in stats if value]
        return cls._wrap_metrics(collected_metrics, tags=tags)

    @staticmethod
    def _get_raw_metrics(url: str, certs: Tuple[str, str]) -> Dict[str, Any]:
        """
        Tries to connect to a K8s Stats Service, receive its response and cast
        it to a dict.

        If for any reason it wasn't able to do this, it returns an empty dict.

        Args:
            url: URL of a Stats Service to gather data from.
            certs: Tuple with paths of a client cert and a client key to
                   authorize.
        Returns: Dict containing K8s Stats Service output.
        """
        try:
            response = requests.get(url, cert=certs, verify=False, timeout=5)
        except (requests.RequestException, IOError) as error:
            logger.warning(
                "[K8sCollector] Could not collect data from %s due to %s", url, error
            )
            return {}
        if response.status_code != 200:
            logger.warning(
                "[K8sCollector] K8s returned %s: %s.",
                response.status_code,
                response.text,
            )
            return {}
        try:
            metrics_dict = response.json()
        except json.JSONDecodeError as error:
            logger.warning(
                "[K8sCollector] K8s returned non-JSON response. Error: %s", error
            )
            return {}
        return metrics_dict



"""
chouette.metrics.plugins.TegrastatsCollector

Stats collector for Nvidia Jetson Devices.
"""
# pylint: disable=too-few-public-methods
import logging
import re
from itertools import chain
from subprocess import Popen, PIPE
from typing import Iterator, List

from pydantic import BaseSettings  # type: ignore
from pykka import ActorDeadError  # type: ignore

from chouette_iot._singleton_actor import SingletonActor
from ._collector_plugin import CollectorPlugin
from .messages import StatsRequest, StatsResponse

__all__ = ["TegrastatsCollector"]

logger = logging.getLogger("chouette-iot")


class TegrastatsConfig(BaseSettings):
    """
    Environment variables based configuration.

    It specifies what metrics this plugin should collect and a path to a
    Tegrastats executable.
    By default all metrics are listed here, but it's possible to request
    a subset of them via an environment variable.
    """

    tegrastats_metrics: List[str] = ["ram", "temp"]
    tegrastats_path: str = "/usr/bin/tegrastats"


class TegrastatsCollector(SingletonActor):
    """
    Actor that collects stats from an Nvidia Tegrastats utility.

    NB: Collectors MUST interact with plugins via `tell` pattern.
        `ask` pattern will return None.
    """

    def __init__(self):
        super().__init__()

        config = TegrastatsConfig()
        self.metrics = config.tegrastats_metrics
        self.path = config.tegrastats_path

    def on_receive(self, message):
        """
        On StatsRequest message collects specified metrics and
        sends them back in a StatsResponse message.

        On any other message does nothing.

        Args:
            message: Expected to be a StatsRequest message.
        """
        logger.debug("[%s] Received %s.", self.name, message)
        if isinstance(message, StatsRequest):
            stats = TegrastatsPlugin.collect_stats(self.path, self.metrics)
            if hasattr(message.sender, "tell"):
                try:
                    message.sender.tell(StatsResponse(self.name, stats))
                except ActorDeadError:
                    logger.warning(
                        "[%s] Requester is stopped. Dropping message.", self.name
                    )


class TegrastatsPlugin(CollectorPlugin):
    """
    CollectorPlugin that handles RAM and Temperature metrics from Tegrastats.

    Built around Tegrastats application that is a part of L4T Nvidia project:
    https://docs.nvidia.com/jetson/l4t/index.html
    """

    @classmethod
    def collect_stats(cls, path: str, metrics_to_collect: List[str]) -> Iterator:
        """
        Collects requested stats from Tegrastats.

        Args:
            path: Path to a Tegrastats executable.
            metrics_to_collect: List of metric types to collect.
        Returns: Iterator over WrappedMetric objects.
        """
        tegrastats_methods = {
            "ram": cls._get_ram_metrics,
            "temp": cls._get_temp_metrics,
        }
        raw_string = cls._get_raw_metrics_string(path)
        methods = filter(None, map(tegrastats_methods.get, metrics_to_collect))
        metrics = [method(raw_string) for method in methods]
        return chain.from_iterable(metrics)

    @classmethod
    def _get_temp_metrics(cls, raw_string: str) -> Iterator:
        """
        Gets temperature of different zones from Tegrastats output.

        Every zone is being sent as a tag for a 'temperature' metric.
        PMIC zone that always shows 100C is dropped.

        Args:
            raw_string: Tegrastats output as a raw sting.
        Returns: Iterator over WrappedMetric objects.
        """
        pattern = re.compile(r"\b(\w+)@([0-9.]+)C\b")
        stats = re.findall(pattern, raw_string)
        metrics = [
            cls._wrap_metrics(
                [("Chouette.tegrastats.temperature", float(value))],
                tags=[f"zone:{zone}"],
            )
            for zone, value in stats
            if zone != "PMIC"
        ]
        return chain.from_iterable(metrics)

    @classmethod
    def _get_ram_metrics(cls, raw_string: str) -> Iterator:
        """
        Gets RAM stats from Tegrastats output.

        It's expected to be less precise than analogous statistics from
        the HostCollectorPlugin, because Tegrastats returns values in MBs
        and to get values in bytes we need to convert these MBs into bytes.

        Args:
            raw_string: Tegrastats output as a raw sting.
        Returns: Iterator over WrappedMetric objects.
        """
        pattern = re.compile(r"\bRAM\b (\d+)/(\d+)MB")
        data = re.findall(pattern, raw_string)
        if not data:
            return iter([])
        used, total = data.pop()
        used_bytes = float(used) * 1024 * 1024
        free_bytes = (float(total) - float(used)) * 1024 * 1024
        collecting_metrics = [
            ("Chouette.tegrastats.ram.used", used_bytes),
            ("Chouette.tegrastats.ram.free", free_bytes),
        ]
        return cls._wrap_metrics(collecting_metrics)

    @staticmethod
    def _get_raw_metrics_string(path: str) -> str:
        """
        Runs Tegrastats and gets a single raw string with stats.

        Args:
            path: Path to a Tegrastats executable.
        Returns: String with raw metrics.
        """
        try:
            ts_proc = Popen(path, stdout=PIPE)
            stdout = ts_proc.stdout
            if stdout:
                data_string = stdout.readline().decode()
            else:
                data_string = ""  # pragma: no cover
            ts_proc.kill()
        except (FileNotFoundError, PermissionError):
            data_string = ""
        return data_string



"""
chouette.metrics.wrappers
"""
# pylint: disable=too-few-public-methods
from typing import Optional

from ._metrics_wrapper import MetricsWrapper
from ._simple_wrapper import SimpleWrapper
from ._datadog_wrapper import DatadogWrapper

__all__ = ["SimpleWrapper", "WrappersFactory", "DatadogWrapper"]


class WrappersFactory:
    """
    WrapperFactory class creates Metrics Wrapper instances.
    """

    wrapper_classes = {"simple": SimpleWrapper, "datadog": DatadogWrapper}

    @classmethod
    def get_wrapper(cls, wrapper_name: str) -> Optional[MetricsWrapper]:
        """
        Takes a wrapper name and returns a wrapper instance.

        Args:
            wrapper_name: Name of a wrapper as a string.
        Returns: ActorRef or None.
        """
        wrapper_class = cls.wrapper_classes.get(wrapper_name)
        wrapper = wrapper_class() if wrapper_class else None
        return wrapper



"""
Metrics classes to handle metrics processing.
"""
# pylint: disable=too-few-public-methods
import time
from typing import Any, Dict, List, Optional

__all__ = ["MergedMetric", "RawMetric", "WrappedMetric"]


class Metric:
    """
    Base parent class for all the metrics.
    """

    __slots__ = ["metric", "type", "tags"]

    def asdict(self):  # pragma: no cover
        """
        Returns a dict form of the metric that is ready to be cast
        to JSON and stored for releasing.

        Return: Dict that represents the metric.
        """
        raise NotImplementedError("Use a concrete Metric class.")

    def __str__(self):
        return str(self.asdict())

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.__str__()}>"

    def __eq__(self, other: object) -> bool:
        """
        Metrics are considered equal if their dicts are equal.

        Args:
            other: Another SingleMetric object to compare.
        Return: Whether their dicts are equal.
        """
        if isinstance(other, Metric):
            return self.asdict() == other.asdict()
        return False


class MergedMetric(Metric):
    """
    MergedMetrics is a single object, that contains aggregated values and
    timestamps of a number of metrics of the same name, type and tags,
    collected during the same `aggregate_interval` period.

    This aggregated data is a raw data that is being transformed by a
    MetricsWrapper object to create WrappedMetrics, ready for release to
    Datadog.

    MergedMetrics are being generated from raw metrics, so they are closer
    to RawMetrics than to WrappedMetrics.

    Like RawMetrics, they have separated `values` and `timestamps` fields
    and their tags are dicts. But when a MergedMetric is generated, it
    creates a `s_tags` property that represents the same tags in a form
    that should be passed to a WrappedMetric - as a list of strings.
    `asdict` method also returns a dict with these processed tags.

    Self.id is a unique identifier of a metric combined of its name, type
    and tags. Only MergedMetrics with the same id can be merged together.
    """

    __slots__ = ["values", "timestamps", "id", "s_tags", "interval"]

    def __init__(self, **kwargs: Any):
        self.metric: str = kwargs["metric"]
        self.type: str = kwargs["type"]
        self.values: List[Any] = kwargs.get("values", [])
        self.timestamps: List[float] = kwargs.get("timestamps", [])
        self.tags: Dict[str, str] = kwargs.get("tags", {})
        self.s_tags: List[str] = self._stringify_tags(self.tags)
        self.id: str = f"{self.metric}_{self.type}{'_'.join(self.s_tags)}"
        self.interval = kwargs.get("interval", 10)

    @staticmethod
    def _stringify_tags(tags: Dict[str, str]) -> List[str]:
        """
        Takes a dict of tags and casts every dict into a line
        "key:value".

        Returns: List of strings with reformatted tags.
        """
        try:
            tags_list = [f"{name}:{str(value)}" for name, value in tags.items()]
        except (AttributeError, TypeError, ValueError):
            tags_list = []
        return sorted(tags_list)

    def __add__(self, other: "MergedMetric"):
        """
        That's the Merge operation of a MergedMetric.

        If there are two MergedMetrics with the same name and same tags, we
        need to be able to merge them and to receive a Metric with the same
        type, same tags and same type whose values and timestamps are merged
        values and timestamps of the original metrics.

        Args:
            other: MergedMetric object to merge with this metric.
        Returns: A new metric with merged values and timestamps.
        """
        if self.id != other.id:
            raise ValueError("Can't merge different metrics.")
        return MergedMetric(
            metric=self.metric,
            type=self.type,
            values=self.values + other.values,
            timestamps=self.timestamps + other.timestamps,
            tags=self.tags,
        )

    def asdict(self):
        """
        Returns: Dict representation of the metric.
        """
        return {
            "metric": self.metric,
            "tags": self.s_tags,
            "values": self.values,
            "timestamps": self.timestamps,
            "type": self.type,
            "interval": self.interval,
        }


class SingleMetric(Metric):
    """
    Parent class for both Wrapped and Raw metrics.
    """

    __slots__ = ["value", "timestamp"]

    def __init__(self, **kwargs: Any):
        self.metric = kwargs["metric"]
        self.type = kwargs["type"]
        self.value = kwargs["value"]
        timestamp = kwargs.get("timestamp")
        self.timestamp = timestamp if timestamp else time.time()


class WrappedMetric(SingleMetric):
    """
    WrappedMetric is a metric that is ready to be released.

    Its timestamp and value are calculated and they form the only data
    point that this metric contains. It usually represents a calculated
    value of some metric for some period of time.

    One MergedMetric, being processed by a MetricsWrapper can produce
    a number of WrappedMetrics.

    Unlike RawMetric or MergedMetric, WrappedMetric's tags are not a dict
    but a list:
    Tags {"tag1": "value1", "tag2": "value2"} should be merged into a list
    ["tag1:value1", "tag2:value2"] before creating a WrappedMetric.

    For aggregated metrics this job is being performed by a MetricsWrapper,
    but for WrappedMetrics produced by CollectionPlugins it should be done
    by a developer.
    """

    __slots__ = ["interval"]

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        tags = kwargs.get("tags")
        self.tags: List[str] = sorted(tags) if tags else []
        self.interval: Optional[int] = kwargs.get("interval")

    def asdict(self):
        """
        Returns a dict form of the metric that is ready to be cast
        to JSON and stored for releasing.

        Return: Dict that represents the metric.
        """
        dict_representation = {
            "metric": self.metric,
            "tags": self.tags,
            "points": [[self.timestamp, self.value]],
            "type": self.type,
        }
        if self.interval:
            dict_representation.update({"interval": self.interval})
        return dict_representation

    def __hash__(self):
        """
        That's a dirty hack needed to avoid duplicated WrappedMetrics for
        HostStatsCollector.

        Return: Hash of a string representation of the metric.
        """
        return hash(self.__str__())


class RawMetric(SingleMetric):
    """
    Raw metric is a metric that needs to be stored to a raw metrics queue
    to be processed on the next aggregator run.

    It's used to store self metrics and represents a single metric datapoint.
    """

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        tags = kwargs.get("tags")
        self.tags: Dict[str, str] = tags if tags else {}

    def asdict(self):
        """
        Returns a dict form of the metric that is ready to be cast
        to JSON and stored for releasing.

        Return: Dict that represents the metric.
        """
        return {
            "metric": self.metric,
            "tags": self.tags,
            "timestamp": self.timestamp,
            "value": self.value,
            "type": self.type,
        }



"""
chouette.metrics.plugins.DockerCollector
"""
# pylint: disable=too-few-public-methods
import json
import logging
import urllib.parse
from concurrent.futures import ThreadPoolExecutor, wait
from itertools import chain
from typing import Iterator, List

import requests_unixsocket  # type: ignore
from pydantic import BaseSettings  # type: ignore
from pykka import ActorDeadError  # type: ignore
from requests import RequestException

from chouette_iot._singleton_actor import SingletonActor
from chouette_iot.metrics import WrappedMetric
from ._collector_plugin import CollectorPlugin
from .messages import StatsRequest, StatsResponse

__all__ = ["DockerCollector"]

logger = logging.getLogger("chouette-iot")


class DockerCollectorConfig(BaseSettings):
    """
    Optional configuration that specifies a path to a docker socket in
    your container.
    """

    docker_socket_path = "/var/run/docker.sock"


class DockerCollector(SingletonActor):
    """
    Docker collector plugin:
    Collects data about docker containers directly from a docker unix socket.
    Should be used if docker-compose or pure docker is used to run software
    on your device.

    NB: Docker socket must be the host docker socket. It must be added as a
    volume to Chouette container.

    Docker-compose file volume addition example:
    volumes:
      - /var/run/docker.sock:/var/run/docker.sock

    Default socket path is /var/run/docker.sock
    """

    def __init__(self):
        super().__init__()
        socket_path = DockerCollectorConfig().docker_socket_path
        encoded_socket_path = urllib.parse.quote(socket_path, safe="")
        self.docker_url = f"http+unix://{encoded_socket_path}/containers"

    def on_receive(self, message: StatsRequest) -> None:
        """
        On StatsRequest message collects Docker statistics and
        sends them back in a StatsResponse message.

        On any other message does nothing.

        Args:
            message: Expected to be a StatsRequest message.
        """
        logger.debug("[%s] Received %s.", self.name, message)
        if isinstance(message, StatsRequest):
            stats = DockerCollectorPlugin.collect_metrics(self.docker_url)
            if hasattr(message.sender, "tell"):
                try:
                    message.sender.tell(StatsResponse(self.name, stats))
                except ActorDeadError:
                    logger.warning(
                        "[%s] Requester is stopped. Dropping message.", self.name
                    )


class DockerCollectorPlugin(CollectorPlugin):
    """
    Companion object for DockerCollector Actor to abstract data collection
    methods into a separated object not handled by Pykka.
    """

    @classmethod
    def collect_metrics(cls, docker_url: str) -> Iterator[WrappedMetric]:
        """
        Facade function of DockerCollectorPlugin.

        Takes a docker URL that includes path to a docker socket file and
        returns an iterator over collected metrics.

        Since there is no single stats endpoint for all the containers, this
        method uses ThreadPoolExecutor to request data about all containers
        in parallel.

        Args:
            docker_url: Docker URL for requests-unixsocket library.
        Returns: Iterator over WrappedMetric objects.
        """
        ids = cls._get_containers_ids(docker_url)
        if not ids:
            return iter([])
        # Generating futures:
        with ThreadPoolExecutor(max_workers=len(ids)) as executor:
            futures = [
                executor.submit(cls._get_container_stats, container_id, docker_url)
                for container_id in ids
            ]
        # Waiting till all the futures are done:
        wait(futures)
        stats = [future.result() for future in futures]

        return chain.from_iterable(stats)

    @classmethod
    def _get_containers_ids(cls, docker_url: str) -> List[str]:
        """
        Connects to a docker socket file, gets a list of all running
        containers and extracts their ids for individual container stats
        requests.

        Args:
            docker_url: Docker URL for requests-unixsocket library.
        Returns: List of container Ids.
        """
        try:
            containers = requests_unixsocket.get(f"{docker_url}/json").json()
        except (TypeError, RequestException, json.JSONDecodeError, IOError) as error:
            logger.warning(
                "[DockerCollector]: Could not get a list of containers due to: %s",
                error,
            )
            return []
        ids = [container["Id"] for container in containers]
        return ids

    @classmethod
    def _get_container_stats(
        cls, container_id: str, docker_url: str
    ) -> Iterator[WrappedMetric]:
        """
        Connects to a docker socket file and collects stats about a specified
        container.
        Usually this process takes around a second, so that's a blocking
        action.
        `stream=false` parameter is used to specify that we just want to get
        a single stats observation without being connected to the socket all
        the time and receiving actual information as a stream. Without it
        DockerCollector would block forever on it first successful stats
        request.

        Args:
            docker_url: Docker URL for requests-unixsocket library.
        Returns: Iterator over WrappedMetric objects.
        """
        try:
            raw_stats = requests_unixsocket.get(
                f"{docker_url}/{container_id}/stats?stream=false"
            ).json()
        except (TypeError, RequestException, json.JSONDecodeError, IOError):
            logger.warning(
                "[DockerCollector]: Could not get stats for a container %s.",
                container_id,
                exc_info=True,
            )
            return iter([])
        # Extracting data from raw stats:
        container_name = raw_stats["name"][1:]
        memory = raw_stats.get("memory_stats", {})
        cpu = raw_stats.get("cpu_stats", {}).get("cpu_usage", {})
        # Generating metrics:
        tags = [f"container:{container_name}"]
        stats = [
            ("Chouette.docker.memory.usage", memory.get("usage")),
            ("Chouette.docker.cpu.usage", cpu.get("usage_in_kernelmode")),
        ]
        # Filtering possible Nones:
        collected_metrics = [(name, value) for name, value in stats if value]
        return cls._wrap_metrics(collected_metrics, tags=tags)



"""
chouette.ChouetteConfig
"""
from typing import List

from pydantic import BaseSettings  # type: ignore

__all__ = ["ChouetteConfig"]


class ChouetteConfig(BaseSettings):
    """
    Main application configuration.

    Data is being collected from environment variables.
    It's case-insensitive, so API_KEY variable will be
    loaded as an api_key parameter.
    """

    api_key: str
    global_tags: List[str]
    collector_plugins: List[str] = []
    aggregate_interval: int = 10
    capture_interval: int = 30
    datadog_url: str = "https://api.datadoghq.com/api"
    log_level: str = "INFO"
    metrics_bulk_size: int = 10000
    metric_ttl: int = 14400
    metrics_wrapper: str = "datadog"
    release_interval: int = 60
    send_self_metrics: bool = True



"""
chouette.Scheduler object for Gevent-based systems.

It won't work without patching time.sleep with gevent.sleep.
"""
import logging
import time
from threading import Lock, Timer
from typing import Any, Callable, Optional, Set

__all__ = ["Scheduler", "Cancellable"]

logger = logging.getLogger("chouette-iot")


class Cancellable:
    """
    Signifies a delayed task that can be cancelled.

    Built around a `Timer` object and is able to cancel it.
    For periodical jobs its `_timer` property is being updated on
    every run from a separated thread, so to avoid race conditions where
    in the middle of the `cancel` run another thread updates the `_timer`
    property with a non-cancelled timer, Lock is used.
    """

    def __init__(self, timer: Optional[Timer]) -> None:
        self._cancelled: bool = False
        self._timer: Optional[Timer] = timer
        self._timer_lock: Lock = Lock()

    def is_cancelled(self) -> bool:
        """
        `._cancelled` property getter for external users.

        There is no external setter for this property.

        Returns: Bool that shows whether it is cancelled.
        """
        return self._cancelled

    def set_timer(self, timer: Timer) -> bool:
        """
        `_.timer` property setter to update timers for periodic jobs.

        There is no external getter for this property.

        Cancelled Cancellable objects shouldn't allow to update their timers.

        Args:
            timer (Timer): Timer object handling a delayed task execution.
        Returns: Bool that shows whether the '_.timer` property was updated.
        """
        with self._timer_lock:
            if self.is_cancelled():
                logger.error(
                    "An attempt to update a timer for a stopped Cancellable happened."
                )
                return False
            self._timer = timer
            return True

    def cancel(self) -> bool:
        """
        Cancels the Cancellable by stopping its timer.

        Returns True only if it is not in a cancelled state.
        If it was requested to cancel an already cancelled Cancellable,
        it returns False to be consistent with a Cancellable object
        from Akka.

        Returns: Bool that shows whether this request actually cancelled
                 the Cancellable timer.
        """
        with self._timer_lock:
            if isinstance(self._timer, Timer):
                self._timer.cancel()
            if self.is_cancelled():
                return False
            self._cancelled = True
            return True


class Scheduler:
    """
    A basic Scheduler service based on Akka Scheduler behaviour.

    Its main purpose is to execute a specified actor once or periodically
    after some delay.
    """

    timers: Set[Cancellable] = set()

    @classmethod
    def stop_all(cls) -> None:
        """
        Stops all the cancellables in its timers set.
        """
        for cancellable in list(cls.timers):
            cancellable.cancel()
            cls.timers.remove(cancellable)

    @classmethod
    def schedule_once(cls, delay: float, func: Callable, *args: Any) -> Cancellable:
        """
        Takes a Callable function and its arguments and creates a timer to run
        it after `delay` seconds.

        Args:
            delay: How many seconds Scheduler waits before executing a func.
            func: Callable that must be executed.
            args: Arguments for the function provided as func.
        Returns: Cancellable object.
        """
        timer = Timer(interval=delay, function=func, args=args)
        timer.start()
        cancellable = Cancellable(timer)
        cls.timers.add(cancellable)
        return cancellable

    @classmethod
    def schedule_at_fixed_rate(
        cls, initial_delay: float, interval: float, func: Callable, *args: Any
    ) -> Cancellable:
        """
        Takes a Callable function and its arguments and creates a timer to be
        run every `interval` seconds.
        This method is precise, so on every run it will try compensate time
        drift to keep the interval between function executions precise.

        Args:
            initial_delay: How many seconds Scheduler waits before the first
                           func execution.
            interval: How many seconds must pass between periodic executions.
            func: Callable that must be executed.
            args: Arguments for the function provided as func.
        Returns: Cancellable object.
        """
        return cls._execute_periodically(
            initial_delay, interval, func, *args, precise=True
        )

    @classmethod
    def schedule_with_fixed_delay(
        cls, initial_delay: float, delay: float, func: Callable, *args: Any
    ) -> Cancellable:
        """
        Takes a Callable function and its arguments and creates a timer to be
        run every `interval` seconds.
        This method is imprecise, so it doesn't try to compensate any time
        drift between calls, so intervals between func executions can be
        different and normally they are a bit longer than specified.

        Args:
            initial_delay: How many seconds Scheduler waits before the first
                           func execution.
            delay: How many seconds must pass between periodic executions.
            func: Callable that must be executed.
            args: Arguments for the function provided as func.
        Returns: Cancellable object.
        """
        return cls._execute_periodically(
            initial_delay, delay, func, *args, precise=False
        )

    @classmethod
    def _execute_periodically(
        cls,
        initial_delay: float,
        interval: float,
        func: Callable,
        *args: Any,
        precise: bool,
    ) -> Cancellable:
        """
        A generic method to handle both precise and imprecise versions of
        periodical func executions.

        It returns a Cancellable object, that can anytime cancel scheduled
        executions by calling its `.cancel()` method. Its `._timer` property
        is being updated via a pseudo-callback `_execute_and_update_cancellable`
        that sends a message and creates a new Timer for another execution.
        To keep it cancellable, this callback replaces an expired timer of
        the returned Cancellable object with this newly created timer.

        Args:
            initial_delay (float): How many seconds we wait before the first execution.
            interval (float): How many seconds we wait between executions.
            func: Callable that must be executed.
            args: Arguments for the function provided as func.
            precise (bool): Defines whether the time drift should be compensated to keep
                            the fixed messaging rate.
        Returns: A Cancellable object.
        """
        cancellable = Cancellable(None)

        started = time.time() + initial_delay

        timer = Timer(
            interval=initial_delay,
            function=cls._execute_and_update_cancellable,
            args=(cancellable, interval, func, *args),
            kwargs={"started": started, "precise": precise},
        )
        timer.start()
        cancellable.set_timer(timer)
        cls.timers.add(cancellable)
        return cancellable

    @classmethod
    def _execute_and_update_cancellable(
        cls,
        cancellable: Cancellable,
        interval: float,
        func: Callable,
        *args: Any,
        started: float,
        precise: bool,
    ) -> None:
        """
        A pseudo-callback function that creates a new Timer for the next
        func execution and updates a Cancellable object with this Timer
        to keep the scheduled activity cancellable.

        If `started` parameter is not None, it tried to be as precise as
        possible and to calculate and compensate time drift from the 'ideal'
        execution time.

        Args:
            cancellable (Cancellable): A previously returned object whose
                `_timer` property we need to update.
            interval (float): How many seconds we wait between executions.
            func: Callable that must be executed.
            args: Arguments for the function provided as func.
            started (Optional[float]): Unix timestamp that says when our
                first execution was happened. If set, it's used to calculate
                and compensate time drift between executions.
        Returns: None, since that's a pseudo-callback.
        """
        now = time.time()
        if precise or now < started:
            time_drift = (now - started) % interval
            delay = interval - time_drift
        else:
            delay = interval

        try:
            if now > started:
                func(*args)
            timer = Timer(
                interval=delay,
                function=cls._execute_and_update_cancellable,
                args=(cancellable, interval, func, *args),
                kwargs={"started": started, "precise": precise},
            )
            timer.start()
            cancellable.set_timer(timer)
        except Exception:
            logger.error("Stopping periodic job because of exception.", exc_info=True)



"""
MetricsCollector class.
"""
import logging
from typing import Any

from pykka import ActorRef, ActorRegistry  # type: ignore

from chouette_iot import ChouetteConfig
from chouette_iot._singleton_actor import VitalActor
from chouette_iot.metrics.plugins import PluginsFactory
from chouette_iot.metrics.plugins.messages import StatsRequest, StatsResponse
from chouette_iot.storages import RedisStorage
from chouette_iot.storages.messages import StoreRecords

logger = logging.getLogger("chouette-iot")

__all__ = ["MetricsCollector"]


class MetricsCollector(VitalActor):
    """
    Actor that is responsible for collecting various stats from a host
    and to store gathered data to a storage for later releasing.
    """

    def __init__(self):
        """
        On creation MetricsCollector reads a list of its plugins from
        environment variables.
        """
        super().__init__()
        config = ChouetteConfig()
        self.plugins = config.collector_plugins
        logger.info(
            "[%s] Starting. Configured collection plugins are: '%s'.",
            self.name,
            "', '".join(self.plugins),
        )

    def on_receive(self, message: Any) -> None:
        """
        On any message that is not a StatResponse one, MetricsCollector
        iterates over its plugins ActorRefs and sends them a StatsRequest
        message.

        They are expected to respond with a StatsResponse message.
        On this message MetricsCollector sends a request to a storage to store
        received metrics.

        Args:
            message: Can be anything.
        """
        if isinstance(message, StatsResponse):
            sender = message.producer
            logger.info("[%s] Storing collected stats from '%s'.", self.name, sender)
            redis = RedisStorage.get_instance()
            redis.tell(StoreRecords("metrics", message.stats, wrapped=True))
        else:
            plugins = map(PluginsFactory.get_plugin, self.plugins)
            for plugin in filter(None, plugins):  # type: ActorRef
                logger.info(
                    "[%s] Requesting stats from '%s'.",
                    self.name,
                    plugin.actor_class.__name__,
                )
                plugin.tell(StatsRequest(self.actor_ref))



import setuptools  # type: ignore

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="chouette-iot",
    version="0.0.2",
    author="Artem Katashev",
    author_email="aharr@rowanleaf.net",
    description="Monitoring and metrics collecting Datadog integration for IoT devices",
    license="Apache License, Version 2.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/akatashev/chouette-iot",
    packages=setuptools.find_packages(),
    install_requires=[
        "redis",
        "pykka",
        "requests",
        "requests-unixsocket",
        "python-json-logger",
        "psutil",
        "pydantic",
        "chouette-iot-client"
    ],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: System :: Monitoring",
    ],
    python_requires=">=3.6",
)



"""
SingletonActor class.
"""
import logging

from pykka import ActorRef, ActorRegistry, ThreadingActor  # type: ignore
from chouette_iot import Scheduler

__all__ = ["SingletonActor", "VitalActor"]

logger = logging.getLogger("chouette-iot")


class SingletonActor(ThreadingActor):
    """
    SingletonActor is a wrapper around pykka actor objects.

    In Chouette workflow we normally use just one instance of an actor.
    SingletonActor is able to return an ActorRef of a running instance
    of its class or to start a new instance and return its ActorRef.
    """

    def __init__(self):
        super().__init__()
        self.name = self.__class__.__name__

    @classmethod
    def get_instance(cls) -> ActorRef:
        """
        Returns a running instance of an actor from ActorRegistry or
        starts a new instance of an actor.

        Returns: ActorRef.
        """
        instances = ActorRegistry.get_by_class(cls)
        if instances:
            return instances.pop()
        return cls.start()

    def on_failure(
        self, exception_type: str, exception_value: str, traceback
    ) -> None:  # pragma: no cover
        """
        Logs an exception if the actor is crashed.

        Args:
            exception_type: Exception type as a string.
            exception_value: Exception value as a string.
            traceback: Traceback object.
        Returns: None.
        """
        logger.error(
            "[%s] Stopped with exception: %s: %s.",
            self.name,
            exception_type,
            exception_value,
            exc_info=True,
        )


class VitalActor(SingletonActor):
    """
    This class represents an actor, that can't be simply restarted.
    If it stopped, the application is stopped with a critical error.
    """

    def on_failure(self, exception_type: str, exception_value: str, traceback) -> None:
        """
        Stops all the actors and all the running timers.
        """
        super().on_failure(exception_type, exception_value, traceback)
        logger.critical("[%s] Stopping Chouette.", self.name)
        ActorRegistry.stop_all()
        Scheduler.stop_all()



"""
MetricsSender actor.
"""
import json
import logging
import zlib
from typing import Any, List, Optional

import requests
from chouette_iot_client import ChouetteClient  # type: ignore
from requests.exceptions import RequestException

from chouette_iot import ChouetteConfig
from chouette_iot._singleton_actor import VitalActor
from chouette_iot.storages import RedisStorage
from chouette_iot.storages._redis_messages import GetHashSizes
from chouette_iot.storages.messages import (
    CleanupOutdatedRecords,
    CollectKeys,
    CollectValues,
    DeleteRecords,
)

__all__ = ["MetricsSender"]

logger = logging.getLogger("chouette-iot")


class MetricsSender(VitalActor):
    """
    MetricsSender is an actor that communicates to Datadog.

    Its responsibility is to cleanup outdated metrics, gather actual
    metrics, compress them and dispatch to Datadog API.
    """

    def __init__(self):
        """
        Next configuration is being extracted from chouette_iotConfig:

        * api_key: Datadog API key.
        * bulk_size: How many metrics we want to send in one bulk at max.
            Default value is 10000. Size of compressed chunk of 10000 metrics
            is around 150KBs. Increase it if your device connection is fast.
        * datadog_url: Datadog URL. It has a default value.
        * metric_ttl: Datadog drops outdated metric, so we clean them before
            sending data. This option says how many seconds is considered
            being "outdated". Metrics older than TTL are being dropped.
        * tags: List of global tags to add to every metric. Should have
            something that gives you a chance to understand what device
            send this metrics. E.g.: a 'host' tag.
        * timeout: Maximum HTTPS Request Timeout for a metrics dispatch
            request.
        """
        super().__init__()
        config = ChouetteConfig()
        self.api_key = config.api_key
        self.bulk_size = config.metrics_bulk_size
        self.datadog_url = config.datadog_url
        self.metric_ttl = config.metric_ttl
        self.redis = RedisStorage.get_instance()
        self.send_self_metrics = config.send_self_metrics
        self.tags = config.global_tags
        self.timeout = int(config.release_interval * 0.8)

    def on_receive(self, message: Any) -> bool:
        """
        On any message MetricsSender:

        1. Performs outdated metrics cleanup prior to gathering data.
        2. Gets a bulk of keys from a RedisStorage actor.
        3. Collects metrics and adds global tags to every of them.
        4. Tries to dispatch them as a compressed "series" message.
        5. If they were dispatched successfully - deletes data from Redis.

        To preserve the exact order of actions, MetricsSender intentionally
        communicates to RedisStorage in a blocking manner, via `ask` requests.

        Args:
            message: Can be anything.
        Returns: Whether data was dispatched and cleaned successfully.
        """
        logger.debug("[%s] Cleaning up outdated wrapped metrics.", self.name)
        self.redis = RedisStorage.get_instance()
        self.redis.ask(
            CleanupOutdatedRecords("metrics", ttl=self.metric_ttl, wrapped=True)
        )
        keys = self.collect_keys()
        if not keys:
            logger.info("[%s] Nothing to dispatch.", self.name)
            return True
        metrics = self.collect_metrics(keys)
        dispatched = self.dispatch_to_datadog(metrics)
        if dispatched:
            cleaned_up = self.redis.ask(DeleteRecords("metrics", keys, wrapped=True))
            if not cleaned_up:
                logger.error(
                    "[%s] Metrics were dispatched, but not cleaned up!", self.name
                )
        else:
            logger.warning(
                "[%s] Metrics were neither dispatched, nor cleaned.", self.name
            )
            cleaned_up = False

        return dispatched and cleaned_up

    def add_global_tags(self, b_metric: bytes) -> Optional[dict]:
        """
        Takes a bytes objects that is expected to represent a JSON object,
        casts it to dict and adds global tags to it list of tags.

        Args:
            b_metric: Bytes object representing a metric as a JSON object.
        Returns: Dict representing a metric with updated tags..
        """
        try:
            d_metric = json.loads(b_metric)
        except (TypeError, json.JSONDecodeError):
            return None
        d_metric["tags"] = d_metric.get("tags", []) + self.tags
        return d_metric

    def collect_keys(self) -> List[bytes]:
        """
        Requests a `self.bulk_size` amount of wrapped metrics keys from Redis.

        It returns the oldest keys to retrieve as much data as possible in case
        of a networking outage.

        Returns: List of metrics keys as bytes.
        """
        request = CollectKeys("metrics", amount=self.bulk_size, wrapped=True)
        keys_and_ts = self.redis.ask(request)
        logger.debug("[%s] Collected %s keys.", self.name, len(keys_and_ts))
        return list(map(lambda pair: pair[0], keys_and_ts))

    def collect_metrics(self, keys: List[bytes]) -> List[dict]:
        """
        Gets a list of metrics from Redis, adds global tags to them and prepare
        them to be dispatched to Datadog.

        Args:
            keys: List of metrics keys as bytes.
        Returns: List of prepared to dispatch metrics.
        """
        b_metrics = self.redis.ask(CollectValues("metrics", keys, wrapped=True))
        logger.debug("[%s] Collected %s metrics.", self.name, len(b_metrics))
        return list(filter(None, map(self.add_global_tags, b_metrics)))

    def dispatch_to_datadog(self, metrics: List[dict]) -> bool:
        """
        Dispatches metrics to Datadog as a "series" POST request.

        https://docs.datadoghq.com/api/v1/metrics/#submit-metrics

        1. It takes the list of prepared metrics.
        2. Casts it to a single "series" request.
        3. Compresses it.
        4. Tries to send it to Datadog.

        If Chouette is expected to send self metrics, as a side
        effect, this function sends 3 metrics:
        1. How many messages are queued to be dispatched this
        minute.
        2. How many metrics were send (if they were sent).
        3. How many bytes were sent (if they were sent).

        Args:
            metrics: List of prepared to dispatch metrics.
        Returns: Whether these metrics were accepted by Datadog.
        """
        # Send a 'chouette.queued.metrics' metric.
        if self.send_self_metrics:
            self.store_queue_size()
        series = json.dumps({"series": metrics})
        compressed_message: bytes = zlib.compress(series.encode())
        metrics_num = len(metrics)
        message_size = len(compressed_message)
        logger.info(
            "[%s] Dispatching %s metrics. Sending around %s KBs of data.",
            self.name,
            metrics_num,
            int(message_size / 1024),
        )
        dispatched = self._post_to_datadog(compressed_message)
        if not dispatched:
            return False
        if self.send_self_metrics:
            ChouetteClient.count("chouette.dispatched.metrics.number", metrics_num)
            ChouetteClient.count("chouette.dispatched.metrics.bytes", message_size)
        return True

    def _post_to_datadog(self, message: bytes) -> bool:
        """
        Implements actual HTTPS interaction with Datadog.

        On message 202 Accepted returns True, on any other message or
        RequestsException returns False and logs an error message.

        Arg:
            message: Compressed message to sent.
        Return: Bool that shows whether the message was accepted.
        """
        try:
            dd_response = requests.post(
                f"{self.datadog_url}/v1/series",
                params={"api_key": self.api_key},
                data=message,
                headers={
                    "Content-Type": "application/json",
                    "Content-Encoding": "deflate",
                },
                timeout=self.timeout,
            )
            if not dd_response.status_code == 202:
                logger.error(
                    "[%s] Unexpected response from Datadog: %s: %s",
                    self.name,
                    dd_response.status_code,
                    dd_response.text,
                )
                return False
        except (RequestException, IOError) as error:
            logger.error(
                "[%s] Could not dispatch metrics due to a HTTP error: %s",
                self.datadog_url,
                error,
            )
            return False
        return True

    def store_queue_size(self) -> None:
        """
        Calculates how many metrics are queued to be dispatched on this
        Sender run and stores this data as a raw metric.

        This is a 'gauge' type metric, because we care about the latest
        value and not about the possible sum of values.

        Returns: None.
        """
        queues_sizes = self.redis.ask(GetHashSizes(["chouette:metrics:wrapped.values"]))
        if queues_sizes:
            _, metrics_queue_size = queues_sizes.pop()
            ChouetteClient.gauge("chouette.queued.metrics", metrics_queue_size)



"""
chouette.metrics.plugins.HostStatsCollector
"""
# pylint: disable=too-few-public-methods
import logging
import time
from itertools import chain
from typing import Iterator, List

import psutil  # type: ignore
from pydantic import BaseSettings  # type: ignore
from pykka import ActorDeadError  # type: ignore

from chouette_iot._singleton_actor import SingletonActor
from chouette_iot.metrics import WrappedMetric
from ._collector_plugin import CollectorPlugin
from .messages import StatsRequest, StatsResponse

__all__ = ["HostStatsCollector"]

logger = logging.getLogger("chouette-iot")


class HostCollectorConfig(BaseSettings):
    """
    Optional Environment variables based configuration.

    It specifies what metrics this plugin should collect. By default all
    metrics are listed here, but it's possible to request a subset of them
    via an environment variable.
    """

    # `network` stats also can be collected.
    host_collector_metrics: List[str] = ["cpu", "fs", "la", "ram"]


class HostStatsCollector(SingletonActor):
    """
    Actor that collects host stats like RAM, CPU and HDD usage.

    NB: Collectors MUST interact with plugins via `tell` pattern.
        `ask` pattern will return None.
    """

    def __init__(self):
        super().__init__()

        host_methods = {
            "cpu": HostCollectorPlugin.get_cpu_percentage,
            "fs": HostCollectorPlugin.get_fs_metrics,
            "la": HostCollectorPlugin.get_la_metrics,
            "ram": HostCollectorPlugin.get_ram_metrics,
            "network": HostCollectorPlugin.get_network_metrics,
        }

        metrics_to_send = HostCollectorConfig().host_collector_metrics
        collection_methods = [
            host_methods.get(method.lower()) for method in metrics_to_send
        ]
        self.methods = [method for method in collection_methods if method]

    def on_receive(self, message: StatsRequest) -> None:
        """
        On StatsRequest message collects specified metrics and
        sends them back in a StatsResponse message.

        On any other message does nothing.

        Args:
            message: Expected to be a StatsRequest message.
        """
        logger.debug("[%s] Received %s.", self.name, message)
        if isinstance(message, StatsRequest):
            metrics = map(lambda func: func(), self.methods)
            stats = chain.from_iterable(metrics)
            if hasattr(message.sender, "tell"):
                try:
                    message.sender.tell(StatsResponse(self.name, stats))
                except ActorDeadError:
                    logger.warning(
                        "[%s] Requester is stopped. Dropping message.", self.name
                    )


class HostCollectorPlugin(CollectorPlugin):
    """
    CollectorPlugin that handles CPU, RAM and HDD metrics.

    Built around psutil package: https://psutil.readthedocs.io/en/latest/
    """

    @classmethod
    def get_la_metrics(cls) -> Iterator[WrappedMetric]:
        """
        Gets LA stats via 'getloadavg()' method:
        https://psutil.readthedocs.io/en/latest/#psutil.getloadavg

        Returns a metric "Chouette.host.la" for 1m LA value.
        There is no real reason to send 5m and 15m metrics, but it's possible,
        commented lines should be uncommented for this.

        Returns: Iterator over WrappedMetric objects.
        """
        min_1, min_5, min_15 = psutil.getloadavg()
        m1m = cls._wrap_metrics([("Chouette.host.la", min_1)], tags=["period:1m"])
        # m5m = cls._wrap_metrics([("Chouette.host.la", min_5)], tags=["period:5m"])
        # m15m = cls._wrap_metrics([("Chouette.host.la", min_15)], tags=["period:15m"])
        # return chain(m1m, m5m, m15m)
        return m1m

    @classmethod
    def get_cpu_percentage(cls) -> Iterator[WrappedMetric]:
        """
        Gets CPU percentage stats via 'cpu_percent()' method:
        https://psutil.readthedocs.io/en/latest/#psutil.cpu_percent

        Documentation says that it can return a dummy 0.0 value on
        the first run, so this value is filtered from the output.

        Returns: Iterator over WrappedMetric objects.
        """
        cpu_percentage = psutil.cpu_percent()
        if cpu_percentage != 0.0:
            collecting_metrics = [("Chouette.host.cpu.percentage", cpu_percentage)]
        else:
            collecting_metrics = []
        return cls._wrap_metrics(collecting_metrics)

    @classmethod
    def get_fs_metrics(cls) -> Iterator[WrappedMetric]:
        """
        Gets disks usage stats.

        Collects the list of partitions and passes it to the
        `_process_filesystem` method to get actual metrics.

        See:
        https://psutil.readthedocs.io/en/latest/#psutil.disk_partitions

        Sometimes Docker returns the same partition as being mounted
        to few different mountpoints. To avoid duplicated metrics,
        WrappedMetrics have __hash__ method and set is being used to
        remove duplicates and send a correct number of metrics for every
        device.

        Returns: Iterator over WrappedMetric objects.
        """
        filesystems = psutil.disk_partitions()
        timestamp = time.time()
        mapped = [list(cls._process_filesystem(fs, timestamp)) for fs in filesystems]
        # Duplication removing hack:
        metrics = iter(set(sum(mapped, [])))
        return metrics

    @classmethod
    def _process_filesystem(cls, filesystem, timestamp) -> Iterator[WrappedMetric]:
        """
        Gets specific filesystem disk usage stats.

        Uses `disk_usage` method to get information about used
        and free storage on a specified filesystem.
        Using this data it's possible to calculate total filesystem
        size or used space percentage in a Datadog dashboard itself.

        See:
        https://psutil.readthedocs.io/en/latest/#psutil.disk_usage

        Args:
            filesystem: psutil._common.sdiskpart object.
        Returns: Iterator over WrappedMetric objects.
        """
        tags = [f"device:{filesystem.device}"]
        fs_usage = psutil.disk_usage(filesystem.mountpoint)
        collecting_metrics = [
            ("Chouette.host.fs.used", fs_usage.used),
            ("Chouette.host.fs.free", fs_usage.free),
        ]
        metrics = cls._wrap_metrics(collecting_metrics, tags=tags, timestamp=timestamp)
        return metrics

    @classmethod
    def get_ram_metrics(cls) -> Iterator[WrappedMetric]:
        """
        Gets memory usage stats via `virtual_memory` method.

        Wraps data about used and available physical memory. Using
        this data it's possible to calculate total memory amount and
        memory usage percentage in a Datadog dashboard itself.

        See:
        https://psutil.readthedocs.io/en/latest/#psutil.virtual_memory

        Returns: Iterator over WrappedMetric objects.
        """
        memory = psutil.virtual_memory()
        collecting_metrics = [
            ("Chouette.host.memory.used", memory.used),
            ("Chouette.host.memory.available", memory.available),
        ]
        return cls._wrap_metrics(collecting_metrics)

    @classmethod
    def get_network_metrics(cls) -> Iterator[WrappedMetric]:
        """
        Gets amount of sent and received bytes for all the interfaces but lo.

        Returns: Iterator over WrappedMetric objects.
        """
        interfaces_data = psutil.net_io_counters(pernic=True)
        metrics = [
            cls._process_iface(iface, data)
            for iface, data in interfaces_data.items()
            if iface != "lo"
        ]
        return chain.from_iterable(metrics)

    @classmethod
    def _process_iface(cls, iface: str, data) -> Iterator[WrappedMetric]:
        """
        Generates `bytes.sent` and `bytes.recv` metrics for a specified iface.

        Args:
            iface: Name of an interface for a tag.
            data: `snetio` namedtuple with data about networking.
        Returns: Iterator over WrappedMetric objects.
        """
        tags = [f"iface:{iface}"]
        collecting_metrics = [
            ("Chouette.host.network.bytes.sent", data.bytes_sent),
            ("Chouette.host.network.bytes.recv", data.bytes_recv),
        ]
        return cls._wrap_metrics(collecting_metrics, tags=tags)
